@@ -51,6 +51,7 @@
 #include <libgnome/gnome-macros.h>
 #include <libgnomevfs/gnome-vfs-ops.h>
 #include <libgnomevfs/gnome-vfs-async-ops.h>
+#include <libgnomevfs/gnome-vfs-mime-utils.h>
 #include <libgnomevfs/gnome-vfs-uri.h>
 #include <libgnomevfs/gnome-vfs-utils.h>
 #include <libgnomevfs/gnome-vfs-xfer.h>
@@ -69,6 +70,7 @@
 #include <libnautilus-private/nautilus-sound.h>
 #include <libnautilus/nautilus-bonobo-ui.h>
 #include <libnautilus/nautilus-clipboard.h>
+#include <libnautilus/nautilus-scroll-positionable.h>
 #include <locale.h>
 #include <signal.h>
 #include <stdio.h>
@@ -96,10 +98,12 @@
 #define COMMAND_TIGHTER_LAYOUT 			"/commands/Tighter Layout"
 #define COMMAND_SORT_REVERSED			"/commands/Reversed Order"
 #define COMMAND_CLEAN_UP			"/commands/Clean Up"
+#define COMMAND_KEEP_ALIGNED 			"/commands/Keep Aligned"
 
 #define ID_MANUAL_LAYOUT                        "Manual Layout"
 #define ID_TIGHTER_LAYOUT                       "Tighter Layout"
 #define ID_SORT_REVERSED                        "Reversed Order"
+#define ID_KEEP_ALIGNED 			"Keep Aligned"
 
 typedef struct {
 	NautilusFileSortType sort_type;
@@ -128,6 +132,8 @@ struct FMIconViewDetails
 	const SortCriterion *sort;
 	gboolean sort_reversed;
 
+	NautilusScrollPositionable *positionable;
+	
 	BonoboUIComponent *ui;
 	
 	NautilusAudioPlayerData *audio_player_data;
@@ -226,7 +232,7 @@ fm_icon_view_finalize (GObject *object)
 	}
 
         if (icon_view->details->react_to_icon_change_idle_id != 0) {
-                gtk_idle_remove (icon_view->details->react_to_icon_change_idle_id);
+                g_source_remove (icon_view->details->react_to_icon_change_idle_id);
         }
 
 	/* kill any sound preview process that is ongoing */
@@ -242,18 +248,7 @@ fm_icon_view_finalize (GObject *object)
 static NautilusIconContainer *
 get_icon_container (FMIconView *icon_view)
 {
-	GtkBin *bin;
-
-	g_return_val_if_fail (FM_IS_ICON_VIEW (icon_view), NULL);
-
-	bin = GTK_BIN (icon_view);
-	if (bin->child) {
-		g_return_val_if_fail (NAUTILUS_IS_ICON_CONTAINER (GTK_BIN (icon_view)->child), NULL);
-		
-		return NAUTILUS_ICON_CONTAINER (GTK_BIN (icon_view)->child);
-	} else {
-		return NULL;
-	}
+	return NAUTILUS_ICON_CONTAINER (GTK_BIN (icon_view)->child);
 }
 
 static gboolean
@@ -447,6 +442,16 @@ handle_radio_item (FMIconView *view,
 	}
 }
 
+static void
+list_covers (NautilusIconData *data, gpointer callback_data)
+{
+	GSList **file_list;
+
+	file_list = callback_data;
+
+	*file_list = g_slist_prepend (*file_list, data);
+}
+
 static void
 unref_cover (NautilusIconData *data, gpointer callback_data)
 {
@@ -457,6 +462,7 @@ static void
 fm_icon_view_clear (FMDirectoryView *view)
 {
 	NautilusIconContainer *icon_container;
+	GSList *file_list;
 	
 	g_return_if_fail (FM_IS_ICON_VIEW (view));
 
@@ -465,8 +471,11 @@ fm_icon_view_clear (FMDirectoryView *view)
 		return;
 
 	/* Clear away the existing icons. */
-	nautilus_icon_container_for_each (icon_container, unref_cover, NULL);
+	file_list = NULL;
+	nautilus_icon_container_for_each (icon_container, list_covers, &file_list);
 	nautilus_icon_container_clear (icon_container);
+	g_slist_foreach (file_list, (GFunc)unref_cover, NULL);
+	g_slist_free (file_list);
 }
 
 
@@ -518,17 +527,23 @@ fm_icon_view_add_file (FMDirectoryView *view, NautilusFile *file)
 	    !should_show_file_on_screen (view, file)) {
 			return;
 	}
-	
+
 	/* Reset scroll region for the first icon added when loading a directory. */
 	if (icon_view->details->loading && nautilus_icon_container_is_empty (icon_container)) {
 		nautilus_icon_container_reset_scroll_region (icon_container);
 	}
+	
 	if (nautilus_icon_container_add (icon_container,
 					 NAUTILUS_ICON_CONTAINER_ICON_DATA (file))) {
 		nautilus_file_ref (file);
 	}
 }
 
+static void
+fm_icon_view_flush_added_files (FMDirectoryView *view)
+{
+	nautilus_icon_container_layout_now (get_icon_container (FM_ICON_VIEW (view)));
+}
 
 static void
 fm_icon_view_file_changed (FMDirectoryView *view, NautilusFile *file)
@@ -565,6 +580,16 @@ fm_icon_view_supports_auto_layout (FMIconView *view)
 		 supports_auto_layout, (view));
 }
 
+static gboolean
+fm_icon_view_supports_keep_aligned (FMIconView *view)
+{
+	g_return_val_if_fail (FM_IS_ICON_VIEW (view), FALSE);
+
+	return EEL_CALL_METHOD_WITH_RETURN_VALUE
+		(FM_ICON_VIEW_CLASS, view,
+		 supports_keep_aligned, (view));
+}
+
 static void
 update_layout_menus (FMIconView *view)
 {
@@ -604,6 +629,18 @@ update_layout_menus (FMIconView *view)
 	nautilus_bonobo_set_sensitive
 		(view->details->ui, COMMAND_CLEAN_UP, !is_auto_layout);	
 
+
+	nautilus_bonobo_set_hidden (view->details->ui,
+				    COMMAND_KEEP_ALIGNED, 
+				    !fm_icon_view_supports_keep_aligned (view));
+	
+	nautilus_bonobo_set_toggle_state 
+		(view->details->ui, COMMAND_KEEP_ALIGNED, 
+		 nautilus_icon_container_is_keep_aligned (get_icon_container (view)));
+	
+	nautilus_bonobo_set_sensitive
+		(view->details->ui, COMMAND_KEEP_ALIGNED, !is_auto_layout);
+
 	bonobo_ui_component_thaw (view->details->ui, NULL);
 }
 
@@ -740,6 +777,41 @@ fm_icon_view_real_set_directory_sort_reversed (FMIconView *icon_view,
 		 sort_reversed);
 }
 
+static gboolean
+get_default_directory_keep_aligned (void)
+{
+	return TRUE;
+}
+
+static gboolean
+fm_icon_view_get_directory_keep_aligned (FMIconView *icon_view,
+					 NautilusFile *file)
+{
+	if (!fm_icon_view_supports_keep_aligned (icon_view)) {
+		return FALSE;
+	}
+	
+	return  nautilus_file_get_boolean_metadata
+		(file,
+		 NAUTILUS_METADATA_KEY_ICON_VIEW_KEEP_ALIGNED,
+		 get_default_directory_keep_aligned ());
+}
+
+static void
+fm_icon_view_set_directory_keep_aligned (FMIconView *icon_view,
+					 NautilusFile *file,
+					 gboolean keep_aligned)
+{
+	if (!fm_icon_view_supports_keep_aligned (icon_view)) {
+		return;
+	}
+
+	nautilus_file_set_boolean_metadata
+		(file, NAUTILUS_METADATA_KEY_ICON_VIEW_KEEP_ALIGNED,
+		 get_default_directory_keep_aligned (),
+		 keep_aligned);
+}
+
 /* maintainence of auto layout boolean */
 static gboolean default_directory_manual_layout = FALSE;
 
@@ -866,6 +938,14 @@ real_supports_auto_layout (FMIconView *view)
 	return TRUE;
 }
 
+static gboolean
+real_supports_keep_aligned (FMIconView *view)
+{
+	g_return_val_if_fail (FM_IS_ICON_VIEW (view), FALSE);
+
+	return FALSE;
+}
+
 static gboolean
 set_sort_reversed (FMIconView *icon_view, gboolean new_value)
 {
@@ -992,6 +1072,9 @@ fm_icon_view_begin_loading (FMDirectoryView *view)
 	/* Set the sort direction from the metadata. */
 	set_sort_reversed (icon_view, fm_icon_view_get_directory_sort_reversed (icon_view, file));
 
+	nautilus_icon_container_set_keep_aligned
+		(get_icon_container (icon_view), 
+		 fm_icon_view_get_directory_keep_aligned (icon_view, file));
 	nautilus_icon_container_set_tighter_layout
 		(get_icon_container (icon_view), 
 		 fm_icon_view_get_directory_tighter_layout (icon_view, file));
@@ -1273,6 +1356,37 @@ sort_reversed_state_changed_callback (BonoboUIComponent *component,
 	}
 }
 
+static void
+keep_aligned_state_changed_callback (BonoboUIComponent *component,
+				     const char        *path,
+				     Bonobo_UIComponent_EventType type,
+				     const char        *state,
+				     gpointer          user_data)
+{
+	FMIconView *icon_view;
+	NautilusFile *file;
+	gboolean keep_aligned;
+
+	g_assert (strcmp (path, ID_KEEP_ALIGNED) == 0);
+
+	icon_view = FM_ICON_VIEW (user_data);
+
+	if (strcmp (state, "") == 0) {
+		/* State goes blank when component is removed; ignore this. */
+		return;
+	}
+
+	keep_aligned = strcmp (state, "1") == 0 ? TRUE : FALSE;
+
+	file = fm_directory_view_get_directory_as_file (FM_DIRECTORY_VIEW (icon_view));
+	fm_icon_view_set_directory_keep_aligned (icon_view,
+						 file,
+						 keep_aligned);
+						      
+	nautilus_icon_container_set_keep_aligned (get_icon_container (icon_view),
+						  keep_aligned);
+}
+
 static void
 switch_to_manual_layout (FMIconView *icon_view)
 {
@@ -1380,6 +1494,7 @@ fm_icon_view_merge_menus (FMDirectoryView *view)
 	
 	bonobo_ui_component_add_listener (icon_view->details->ui, ID_TIGHTER_LAYOUT, tighter_layout_state_changed_callback, view);
 	bonobo_ui_component_add_listener (icon_view->details->ui, ID_SORT_REVERSED, sort_reversed_state_changed_callback, view);
+	bonobo_ui_component_add_listener (icon_view->details->ui, ID_KEEP_ALIGNED, keep_aligned_state_changed_callback, view);
 	icon_view->details->menus_ready = TRUE;
 
 	bonobo_ui_component_freeze (icon_view->details->ui, NULL);
@@ -1459,6 +1574,8 @@ fm_icon_view_reset_to_defaults (FMDirectoryView *view)
 
 	set_sort_criterion (icon_view, get_sort_criterion_by_sort_type (get_default_sort_order ()));
 	set_sort_reversed (icon_view, get_default_sort_in_reverse_order ());
+	nautilus_icon_container_set_keep_aligned 
+		(icon_container, get_default_directory_keep_aligned ());
 	nautilus_icon_container_set_tighter_layout
 		(icon_container, get_default_directory_tighter_layout ());
 
@@ -1586,7 +1703,7 @@ play_file (gpointer callback_data)
 	file = icon_view->details->audio_preview_file;
 	file_uri = nautilus_file_get_uri (file);
 	mime_type = nautilus_file_get_mime_type (file);
-	is_mp3 = eel_strcasecmp (mime_type, "audio/x-mp3") == 0;
+	is_mp3 = eel_strcasecmp (mime_type, "audio/mpeg") == 0;
 	is_ogg = eel_strcasecmp (mime_type, "application/x-ogg") == 0;
 
 	mp3_pid = fork ();
@@ -1666,7 +1783,7 @@ play_file (gpointer callback_data)
 	file_path = gnome_vfs_get_local_path_from_uri (file_uri);
 	mime_type = nautilus_file_get_mime_type (icon_view->details->audio_preview_file);
 
-	is_mp3 = eel_strcasecmp (mime_type, "audio/x-mp3") == 0;
+	is_mp3 = eel_strcasecmp (mime_type, "audio/mpeg") == 0;
 
 	if (file_path != NULL && !is_mp3) {
 		icon_view->details->audio_player_data = nautilus_audio_player_play (file_path);
@@ -1704,17 +1821,17 @@ preview_audio (FMIconView *icon_view, NautilusFile *file, gboolean start_flag)
 	}
 #endif
 	if (icon_view->details->audio_preview_timeout != 0) {
-		gtk_timeout_remove (icon_view->details->audio_preview_timeout);
+		g_source_remove (icon_view->details->audio_preview_timeout);
 		icon_view->details->audio_preview_timeout = 0;
 	}
 			
 	if (start_flag) {
 		icon_view->details->audio_preview_file = file;
 #if USE_OLD_AUDIO_PREVIEW			
-		icon_view->details->audio_preview_timeout = gtk_timeout_add (1000, play_file, icon_view);
+		icon_view->details->audio_preview_timeout = g_timeout_add (1000, play_file, icon_view);
 #else
 		/* FIXME: Need to kill the existing timeout if there is one? */
-		icon_view->details->audio_preview_timeout = gtk_timeout_add (1000, play_file, icon_view);
+		icon_view->details->audio_preview_timeout = g_timeout_add (1000, play_file, icon_view);
 #endif
 	}
 }
@@ -1820,6 +1937,44 @@ fm_icon_view_filter_by_screen (FMIconView *icon_view,
 	icon_view->details->num_screens = gdk_display_get_n_screens (gtk_widget_get_display (GTK_WIDGET (icon_view)));
 }
 
+static void
+fm_icon_view_screen_changed (GtkWidget *widget,
+			     GdkScreen *previous_screen)
+{
+	FMDirectoryView *view;
+	GList *files, *l;
+	NautilusFile *file;
+	NautilusIconContainer *icon_container;
+
+	if (GTK_WIDGET_CLASS (parent_class)->screen_changed) {
+		GTK_WIDGET_CLASS (parent_class)->screen_changed (widget, previous_screen);
+	}
+
+	view = FM_DIRECTORY_VIEW (widget);
+	if (FM_ICON_VIEW (view)->details->filter_by_screen) {
+		icon_container = get_icon_container (FM_ICON_VIEW (view));
+		
+		files = nautilus_directory_get_file_list (fm_directory_view_get_model (view));
+
+		for (l = files; l != NULL; l = l->next) {
+			file = l->data;
+			
+			if (!should_show_file_on_screen (view, file)) {
+				fm_icon_view_remove_file (view, file);
+			} else {
+				if (nautilus_icon_container_add (icon_container,
+								 NAUTILUS_ICON_CONTAINER_ICON_DATA (file))) {
+					nautilus_file_ref (file);
+				}
+			}
+		}
+		
+		nautilus_file_list_unref (files);
+		g_list_free (files);
+	}
+}
+
+
 static int
 compare_files_cover (gconstpointer a, gconstpointer b, gpointer callback_data)
 {
@@ -1925,8 +2080,8 @@ icon_position_changed_callback (NautilusIconContainer *container,
 	 */
 	if (icon_view->details->react_to_icon_change_idle_id == 0) {
                 icon_view->details->react_to_icon_change_idle_id
-                        = gtk_idle_add (fm_icon_view_react_to_icon_change_idle_callback,
-                                        icon_view);
+                        = g_idle_add (fm_icon_view_react_to_icon_change_idle_callback,
+				      icon_view);
 	}
 
 	/* Store the new position of the icon in the metadata. */
@@ -2286,6 +2441,7 @@ icon_view_handle_uri_list (NautilusIconContainer *container, const char *item_ur
 	char *path;
 	char *stripped_uri;
 	char *container_uri;
+	char *mime_type;
 	const char *last_slash, *link_name;
 	int n_uris;
 	gboolean all_local;
@@ -2295,7 +2451,7 @@ icon_view_handle_uri_list (NautilusIconContainer *container, const char *item_ur
 		return;
 	}
 
-	container_uri = fm_directory_view_get_uri (FM_DIRECTORY_VIEW (view));
+	container_uri = fm_directory_view_get_backing_uri (FM_DIRECTORY_VIEW (view));
 	g_return_if_fail (container_uri != NULL);
 
 	if (eel_vfs_has_capability (container_uri,
@@ -2376,11 +2532,20 @@ icon_view_handle_uri_list (NautilusIconContainer *container, const char *item_ur
 			/* Make a link using the desktop file contents? */
 			uri = node->data;
 			path = gnome_vfs_get_local_path_from_uri (uri);
-			
+
 			if (path != NULL) {
-				entry = gnome_desktop_item_new_from_file (path,
-									  GNOME_DESKTOP_ITEM_LOAD_ONLY_IF_EXISTS,
-									  NULL);
+				mime_type = gnome_vfs_get_mime_type (uri);
+
+				if (mime_type != NULL &&
+				    strcmp (mime_type, "application/x-gnome-app-info") == 0) {
+					entry = gnome_desktop_item_new_from_file (path,
+							GNOME_DESKTOP_ITEM_LOAD_ONLY_IF_EXISTS,
+							NULL);
+				} else {
+					entry = NULL;
+				}
+
+				g_free (mime_type);
 				g_free (path);
 			} else {
 				entry = NULL;
@@ -2422,6 +2587,38 @@ icon_view_handle_uri_list (NautilusIconContainer *container, const char *item_ur
 	
 }
 
+static char *
+icon_view_get_first_visible_file_callback (NautilusScrollPositionable *positionable,
+					   FMIconView *icon_view)
+{
+	NautilusFile *file;
+
+	file = NAUTILUS_FILE (nautilus_icon_container_get_first_visible_icon (get_icon_container (icon_view)));
+
+	if (file) {
+		return nautilus_file_get_uri (file);
+	}
+	
+	return NULL;
+}
+
+static void
+icon_view_scroll_to_file_callback (NautilusScrollPositionable *positionable,
+				   const char *uri,
+				   FMIconView *icon_view)
+{
+	NautilusFile *file;
+
+	if (uri != NULL) {
+		file = nautilus_file_get (uri);
+		if (file != NULL) {
+			nautilus_icon_container_scroll_to_icon (get_icon_container (icon_view),
+								NAUTILUS_ICON_CONTAINER_ICON_DATA (file));
+			nautilus_file_unref (file);
+		}
+	}
+}
+
 static void
 fm_icon_view_class_init (FMIconViewClass *klass)
 {
@@ -2431,7 +2628,10 @@ fm_icon_view_class_init (FMIconViewClass *klass)
 
 	G_OBJECT_CLASS (klass)->finalize = fm_icon_view_finalize;
 	
+	GTK_WIDGET_CLASS (klass)->screen_changed = fm_icon_view_screen_changed;
+	
 	fm_directory_view_class->add_file = fm_icon_view_add_file;
+	fm_directory_view_class->flush_added_files = fm_icon_view_flush_added_files;
 	fm_directory_view_class->begin_loading = fm_icon_view_begin_loading;
 	fm_directory_view_class->bump_zoom_level = fm_icon_view_bump_zoom_level;
 	fm_directory_view_class->can_rename_file = fm_icon_view_can_rename_file;
@@ -2454,9 +2654,9 @@ fm_icon_view_class_init (FMIconViewClass *klass)
 	fm_directory_view_class->zoom_to_level = fm_icon_view_zoom_to_level;
         fm_directory_view_class->click_policy_changed = fm_icon_view_click_policy_changed;
         fm_directory_view_class->embedded_text_policy_changed = fm_icon_view_embedded_text_policy_changed;
+        fm_directory_view_class->emblems_changed = fm_icon_view_emblems_changed;
         fm_directory_view_class->image_display_policy_changed = fm_icon_view_image_display_policy_changed;
         fm_directory_view_class->merge_menus = fm_icon_view_merge_menus;
-        fm_directory_view_class->emblems_changed = fm_icon_view_emblems_changed;
         fm_directory_view_class->sort_directories_first_changed = fm_icon_view_sort_directories_first_changed;
         fm_directory_view_class->start_renaming_file = fm_icon_view_start_renaming_file;
         fm_directory_view_class->text_attribute_names_changed = fm_icon_view_text_attribute_names_changed;
@@ -2464,6 +2664,7 @@ fm_icon_view_class_init (FMIconViewClass *klass)
 
 	klass->clean_up = fm_icon_view_real_clean_up;
 	klass->supports_auto_layout = real_supports_auto_layout;
+	klass->supports_keep_aligned = real_supports_keep_aligned;
         klass->get_directory_auto_layout = fm_icon_view_real_get_directory_auto_layout;
         klass->get_directory_sort_by = fm_icon_view_real_get_directory_sort_by;
         klass->get_directory_sort_reversed = fm_icon_view_real_get_directory_sort_reversed;
@@ -2478,6 +2679,7 @@ static void
 fm_icon_view_instance_init (FMIconView *icon_view)
 {
 	static gboolean setup_sound_preview = FALSE;
+	NautilusView *nautilus_view;
 
         g_return_if_fail (GTK_BIN (icon_view)->child == NULL);
 
@@ -2487,6 +2689,11 @@ fm_icon_view_instance_init (FMIconView *icon_view)
 
 	create_icon_container (icon_view);
 
+	icon_view->details->positionable = nautilus_scroll_positionable_new ();
+	nautilus_view = fm_directory_view_get_nautilus_view (FM_DIRECTORY_VIEW (icon_view));
+	bonobo_object_add_interface (BONOBO_OBJECT (nautilus_view),
+				     BONOBO_OBJECT (icon_view->details->positionable));
+	
 	if (!setup_sound_preview) {
 		eel_preferences_add_auto_enum (NAUTILUS_PREFERENCES_PREVIEW_SOUND,
 					       &preview_sound_auto_value);
@@ -2516,5 +2723,8 @@ fm_icon_view_instance_init (FMIconView *icon_view)
 
 	g_signal_connect_object (get_icon_container (icon_view), "handle_uri_list",
 				 G_CALLBACK (icon_view_handle_uri_list), icon_view, 0);
-
+	g_signal_connect_object (icon_view->details->positionable, "get_first_visible_file",
+				 G_CALLBACK (icon_view_get_first_visible_file_callback), icon_view, 0);
+	g_signal_connect_object (icon_view->details->positionable, "scroll_to_file",
+				 G_CALLBACK (icon_view_scroll_to_file_callback), icon_view, 0);
 }
